/* negotiat.c       Copyright (c) 2000-2002 Nagy Daniel
 *
 * $Date: 2002/03/28 17:02:39 $
 * $Revision: 1.12 $
 *
 * This module is the SSH negotiation part:
 *  - open TCP connection
 *  - version check
 *  - key generation
 *  - user authorization
 *  - TTY allocation
 *  - interactive shell request
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include<io.h>
#include<stdio.h>
#include<stdlib.h>
#include<conio.h>
#include<string.h>

#if defined (__DJGPP__)
 #include<errno.h>
 #include"tcp_djgp.h"
#elif defined (__TURBOC__)
 #include"tcp.h"
#endif

#include"ssh.h"
#include"protocol.h"
#include"macros.h"
#include"md5.h"
#include"rsa.h"
#include"des.h"
#include"blowfish.h"
#include"version.h"
#include"xmalloc.h"

void fatal(const char *fmt, ...);

extern int cipher;		/* cipher of or off*/
extern int cipher_type;         /* type of cipher */
extern int ssh_port;		/* SSH port of server */
extern tcp_Socket s;
extern unsigned char username[20];
extern unsigned char password[MAX_PASSWORD_LENGTH];
extern short passwdcmd;
extern int status;
extern DESCon keys[3];
extern BlowfishContext ectx, dctx;
extern char *term;		/* terminal type */
extern char verbose;		/* verbose mode */

extern struct Packet pktin;
extern struct Packet pktout;

char done[]="Done\r\n";		/* for verbose messages */

char privileged=1; /* Use privileged port by default */

/* Authentication mask bits */
enum { SSH_AUTH_RHOSTS = 1, SSH_AUTH_RSA = 2, SSH_AUTH_PASSWORD = 3,
    SSH_AUTH_RHOSTS_RSA = 4 };
unsigned char auth_mask[4];

/* check_emulation: take the remote party's version number as
   arguments and return our possibly modified version number back
   (relevant only for clients).

   Return values:
   EMULATE_VERSION_OK means we can work together

   EMULATE_VERSION_TOO_OLD if the other party has too old version
   which we cannot emulate,

   EMULATE_MAJOR_VERSION_MISMATCH if the other party has different
   major version and thus will probably not understand anything we
   say, and

   EMULATE_VERSION_NEWER if the other party has never code than we
   have.

   */

static int check_emulation(int remote_major, int remote_minor,
		    int *return_major, int *return_minor)
{
  if (return_major)
    *return_major = PROTOCOL_MAJOR;
  if (return_minor)
    {
      if (remote_minor < PROTOCOL_MINOR)
	*return_minor = remote_minor;
      else
	*return_minor = PROTOCOL_MINOR;
    }

  if (remote_major < PROTOCOL_MAJOR)
    return EMULATE_MAJOR_VERSION_MISMATCH;

  if (remote_major == 1 && remote_minor == 0)
    return EMULATE_VERSION_TOO_OLD;  /* We no longer support 1.0. */
  
  if (remote_major > PROTOCOL_MAJOR ||
      (remote_major == PROTOCOL_MAJOR && remote_minor > PROTOCOL_MINOR))
    {
      /* The remote software is newer than we. If we are the client,
	 no matter - the server will decide. If we are the server, we
	 cannot emulate a newer client and decide to stop. */
      return EMULATE_VERSION_NEWER;
    }

  return EMULATE_VERSION_OK;
}


/* SSH version string exchange: get server's SSH protocol
   version, examine it, and send ours if it seems that we
   can communicate
   */

static void ssh_exchange_identification(void)
{
char buf[256], remote_version[256];
int remote_major, remote_minor, i;
int my_major, my_minor;
int len;

	if(verbose) cputs("Identification Exchange\n\r");

	/* Read other side's version identification. */
	i=sock_gets(&s, buf, sizeof(buf));
	if(!i) fatal("read: %s", strerror(errno));
	buf[i] = '\n';

	if(verbose) cprintf("Remote version: %s\r",buf);

	/* Check that the versions match.  In future this might accept several
	   versions and set appropriate flags to handle them. */
	if (sscanf(buf, "SSH-%d.%d-%[^\n]\n", &remote_major, &remote_minor,
	     remote_version) != 3)
		fatal("Bad remote protocol version identification\n");

	switch (check_emulation(remote_major, remote_minor,
		     &my_major, &my_minor)) {
	    case EMULATE_VERSION_TOO_OLD:
	      fatal("Remote machine has too old SSH software version\n");
	    case EMULATE_MAJOR_VERSION_MISMATCH:
	      fatal("Major protocol versions incompatible\n");
	    case EMULATE_VERSION_NEWER:
	      /* We will emulate the old version. */
	      break;
	    case EMULATE_VERSION_OK:
	      break;
	    default:
	      fatal("Unexpected return value from check_emulation\n");
	} /* switch */

	sprintf(buf, "SSH-%d.%d-%s\n\0", my_major, my_minor, SSH_VERSION);
	if(verbose) cprintf("Local version: %s\r",buf);
	len=sock_write(&s, buf, strlen(buf));
	if(len != strlen(buf)) fatal("write: %s", strerror(errno));
}

/* create session key and ID */
static void create_ssh_keys(void)
{
int i,j,len;
unsigned char session_key[32];
unsigned char session_id[16];
unsigned char *RSAblock, *keystr1, *keystr2;
unsigned char cookie[8];

R_RSAKey servkey, hostkey;
MD5Context md5c;

    /* get the server key */
    memcpy(cookie, pktin.body, 8);
    if(verbose)cputs("Extracting server keys...");
    i=makekey(pktin.body+8, &servkey, &keystr1);
    j=makekey(pktin.body+8+i, &hostkey, &keystr2);
    memcpy(auth_mask, pktin.body+16+i+j, sizeof(auth_mask));
    if(verbose)cputs(done);

    /* generate session ID */
    if(verbose)cputs("Generating session ID...");
    MD5Init(&md5c);
    MD5Update(&md5c, keystr2, hostkey.bytes);
    MD5Update(&md5c, keystr1, servkey.bytes);
    MD5Update(&md5c, cookie, 8);
    MD5Final(session_id, &md5c);
    if(verbose)cputs(done);

    for (i=0; i<32; i++)
	session_key[i] = rand() % 256;

    len = (hostkey.bytes > servkey.bytes ? hostkey.bytes : servkey.bytes);

    RSAblock = (unsigned char *)xmalloc(len);
    memset(RSAblock,0,len);

    for (i=0; i<32; i++) {
	RSAblock[i] = session_key[i];
	if (i < 16)
	    RSAblock[i] ^= session_id[i];
    }


    if(verbose)cputs("Encrypting session key...");
    if (hostkey.bytes > servkey.bytes) {
	rsaencrypt(RSAblock, 32, &servkey);
	rsaencrypt(RSAblock, servkey.bytes, &hostkey);
    } else {
	rsaencrypt(RSAblock, 32, &hostkey);
	rsaencrypt(RSAblock, hostkey.bytes, &servkey);
    }
    if(verbose)cputs(done);

    s_wrpkt_start(SSH_CMSG_SESSION_KEY, len+15);
    pktout.body[0] = cipher_type;
    memcpy(pktout.body+1, cookie, 8);
    pktout.body[9] = (len*8) >> 8;
    pktout.body[10] = (len*8) & 0xFF;
    memcpy(pktout.body+11, RSAblock, len);
    pktout.body[len+11] = pktout.body[len+12] = 0;
    pktout.body[len+13] = pktout.body[len+14] = 0;
    if(verbose)cputs("Sending encrypted session key...");
    s_wrpkt();
    if(verbose)cputs(done);
    xfree(RSAblock);

    switch(cipher_type)	{
	case SSH_CIPHER_3DES:
		des_set_key(session_key, &keys[0]);
		des_set_key(session_key+8, &keys[1]);
		des_set_key(session_key+16, &keys[2]);
		break;

	case SSH_CIPHER_BLOWFISH:
		blowfish_setkey(&ectx, session_key, SSH_SESSION_KEY_LENGTH);
		    ectx.biv0 = 0;
		    ectx.biv1 = 0;
		    dctx = ectx;
		break;

    } /* switch */

    cipher=1; /* enable cipher */
}

/* Send username */
static void SendUsername(char *usrname)
{
int i;

    if(verbose) cputs("Sending username...");
    i=strlen(usrname);
    s_wrpkt_start(SSH_CMSG_USER, i+4);
    pktout.body[0] = pktout.body[1] = pktout.body[2] = 0;
    pktout.body[3] = i;
    memcpy(pktout.body+4, usrname, i);
    s_wrpkt();
    if(verbose) cputs(done);
}

/* get password */
static void AskPassword(void)
{
short len=0;
unsigned char ch;

	if(passwdcmd) len = strlen(password);
	else{
		cputs("Password: ");
		while((ch=getch()) != 0x0d && len < MAX_PASSWORD_LENGTH){
			if(ch == 8 && len>0)
				password[--len]=0;
			else if(ch >= 32)
				password[len++]=ch;
		}
	}
	cprintf("\n\r");
	s_wrpkt_start(SSH_CMSG_AUTH_PASSWORD, len+4);
	pktout.body[0] = pktout.body[1] = pktout.body[2] = 0;
	pktout.body[3] = len;
	memcpy(pktout.body+4, password, len);
	s_wrpkt();
	memset(password, 0, MAX_PASSWORD_LENGTH); /* security */
}


/* set maximum packet size */
static void packetsize(void)
{
unsigned int size=MAX_PACKET_SIZE;

	if(verbose) cputs("Setting maximum packet size...");
	s_wrpkt_start(SSH_CMSG_MAX_PACKET_SIZE, 4);
	pktout.body[0] = pktout.body[1] = 0;
	pktout.body[2] = size >> 8;
	pktout.body[3] = size & 0xFF;
	s_wrpkt();
	packet_read_block();
	if(verbose) {
	  if(pktin.type==SSH_SMSG_SUCCESS) cputs("success\r\n");
	  else cputs("failed\r\n");
	} /* if */
}

/* Negotiate the connection with server */
int connectssh(char *remotehost)
{
int n;
int localport;
longword remoteip;

    /* Allocate local port */
    localport = (rand() % 512) + 512;
    if(!privileged) localport = localport + 512;

    sock_init(); /* Initalize socket */

    /* Resolve hostname */
    if ((remoteip = resolve(remotehost)) == 0) {
       cprintf("\n\rUnable to resolve `%s'\n\r", remotehost);
       return(1);
    }

/* Open TCP */
    if (!tcp_open(&s, localport, remoteip, ssh_port, NULL)) {
       cputs("Unable to open connection\n\r");
       return(1);
    }

/* Negotiate TCP connection */
    cputs("waiting for remote host to connect...\n\r");
    fflush(stdout);
    sock_wait_established (&s, sock_delay,NULL, &status);

    sock_wait_input (&s, sock_delay, NULL, &status);

/* Begin SSH negotiation protocol */

    /* Start negotiation on network */
    ssh_exchange_identification();
    sock_tick(&s, &status);

    /* Wait for a public key packet from the server. */
    if(verbose) cputs("Wait for public key\n\r");
    packet_read_expect(SSH_SMSG_PUBLIC_KEY);

    /* Create SSH keys */
    create_ssh_keys();

    /* Check that our authentication method is allowed */
    if ((auth_mask[3] & (1 << SSH_AUTH_PASSWORD)) == 0)
	fatal("Server does not allow password authentication");

    /* Wait for key confirmation */
    if(verbose) cputs("Waiting for first encrypted ACK\n\r");
    packet_read_expect(SSH_SMSG_SUCCESS);

    /* Send username */
    SendUsername(username);
    packet_read_block();
    switch(pktin.type) {
	case SSH_SMSG_SUCCESS:		/* no authentication needed */
		break;
	case SSH_SMSG_FAILURE:  	/* get password */
		if(passwdcmd) n=1; /* if password is specified */
		else n=3; /* three chances if not specified yet*/
		while(1)
		{
		 AskPassword();
		 packet_read_block();
		 if(pktin.type==SSH_SMSG_SUCCESS) break;
		 n--;
		 if(n==0) fatal("Invalid password\n");
		 else cputs("Invalid password\n\r");
		}
		break;
	default:
		fatal("Invalid packet received");
	}

    /* Set maximum packet size */
    packetsize();

    return(0);

sock_err:
  switch (status)
  {
    case 1 : cputs ("Connection closed\n\r");
             break;
    case -1: cputs ("REMOTE HOST CLOSED CONNECTION\n\r");
             break;
  }
 return(1);
}
